// PV112 2017, lesson 3

#include <iostream>
#include "PV112.h"

#define _USE_MATH_DEFINES
#include <math.h>

// Include GLEW, use static library
#define GLEW_STATIC
#if defined(_WIN32)
#define NOMINMAX // Make Windows.h not define 'min' and 'max' macros
#include <GL/wglew.h> // Include on Windows
#pragma comment(lib, "glew32s.lib")
#else
#include <GL/glxew.h> // Include on Linux and Mac
#define APIENTRY
#endif

#include <GL/freeglut.h>

#include <glm/glm.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>

using namespace std;
using namespace PV112;

// Current window size
int win_width = 480;
int win_height = 480;

GLuint program;

GLint color_loc;
GLint model_matrix_loc;
GLint view_matrix_loc;
GLint projection_matrix_loc;
GLint PVM_matrix_loc;
GLint normal_matrix_loc;

GLint material_ambient_color_loc;
GLint material_diffuse_color_loc;
GLint material_specular_color_loc;
GLint material_shininess_loc;

GLint light_position_loc;
GLint light_diffuse_color_loc;
GLint light_ambient_color_loc;
GLint light_specular_color_loc;

GLint eye_position_loc;

// We are only going to use teapots
PV112Geometry my_teapot;

// Simple camera that allows us to look at the object from different views
PV112Camera my_camera;

// Current time of the application in seconds, for animations
float app_time_s = 0.0f;

// Called when the user presses a key
void key_pressed(unsigned char key, int mouseX, int mouseY)
{
    switch (key)
    {
    case 27: // Escape
        exit(0);
        break;
    case 'l':
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        glutPostRedisplay();
        break;
    case 'f':
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glutPostRedisplay();
        break;
    case 't':
        glutFullScreenToggle();
        break;
    }
}

// Called when the user presses a mouse button
void mouse_button_changed(int button, int state, int x, int y)
{
    my_camera.OnMouseButtonChanged(button, state, x, y);
}

// Called when the user moves with the mouse (when some mouse button is pressed)
void mouse_moved(int x, int y)
{
    my_camera.OnMouseMoved(x, y);
}

// Initializes OpenGL stuff
void init()
{
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepth(1.0);
    glEnable(GL_DEPTH_TEST);

    // Create shader program
    program = CreateAndLinkProgram("vertex.glsl", "fragment.glsl");
    if (0 == program)
        WaitForEnterAndExit();

    // Get attribute and uniform locations
    int position_loc = glGetAttribLocation(program, "position");
    int normal_loc = glGetAttribLocation(program, "normal");
    color_loc = glGetUniformLocation(program, "color");
    model_matrix_loc = glGetUniformLocation(program, "model_matrix");
    PVM_matrix_loc = glGetUniformLocation(program, "PVM_matrix");
    normal_matrix_loc = glGetUniformLocation(program, "normal_matrix");

    light_position_loc = glGetUniformLocation(program, "light_position");
    light_ambient_color_loc = glGetUniformLocation(program, "light_ambient_color");
    light_diffuse_color_loc = glGetUniformLocation(program, "light_diffuse_color");
    light_specular_color_loc = glGetUniformLocation(program, "light_specular_color");
    eye_position_loc = glGetUniformLocation(program, "eye_position");

    material_ambient_color_loc = glGetUniformLocation(program, "material_ambient_color");
    material_diffuse_color_loc = glGetUniformLocation(program, "material_diffuse_color");
    material_specular_color_loc = glGetUniformLocation(program, "material_specular_color");
    material_shininess_loc = glGetUniformLocation(program, "material_shininess");

    my_teapot = CreateTeapot(position_loc, normal_loc);
}

// Called when the window needs to be rerendered
void render()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glm::mat4 projection_matrix, view_matrix, model_matrix, PVM_matrix;
    glm::mat3 normal_matrix;

    projection_matrix = glm::perspective(glm::radians(45.0f),
            float(win_width) / float(win_height), 0.1f, 100.0f);
    view_matrix = glm::lookAt(my_camera.GetEyePosition(),
            glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 1.0f, 0.0f));

    glUseProgram(program);


    glBindVertexArray(my_teapot.VAO);

    glUniform3f(color_loc, 0.6f, 0.6f, 0.6f);
    model_matrix = glm::mat4(1.0f);

    PVM_matrix = projection_matrix * view_matrix * model_matrix;

    normal_matrix = glm::inverse(glm::transpose(glm::mat3(model_matrix)));

    glUniformMatrix4fv(model_matrix_loc, 1, GL_FALSE, glm::value_ptr(model_matrix));
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    glUniformMatrix3fv(normal_matrix_loc, 1, GL_FALSE, glm::value_ptr(normal_matrix));

    DrawGeometry(my_teapot);


    glBindVertexArray(0);
    glUseProgram(0);

    glutSwapBuffers();
}

// Called when the window changes its size
void reshape(int width, int height)
{
    win_width = width;
    win_height = height;

    // Set the area into which we render
    glViewport(0, 0, win_width, win_height);
}

// Simple timer function for animations
void timer(int)
{
    // Add 20 ms to the application time, register the timer to be called again
    // in 20 ms, and rerender the scene.
    app_time_s += 0.020f;
    glutTimerFunc(20, timer, 0);
    glutPostRedisplay();
}


// Callback function to be called when we make an error in OpenGL
void APIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
    switch (type) // Report only some errors
    {
    case GL_DEBUG_TYPE_ERROR:
    case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
    case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
        cout << message << endl; // Put the breakpoint here
        return;
    default:
        return;
    }
}

int main(int argc, char ** argv)
{
    // Initialize GLUT
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

    // Set OpenGL Context parameters
    glutInitContextVersion(3, 3);              // Use OpenGL 3.3
    glutInitContextProfile(GLUT_CORE_PROFILE); // Use OpenGL core profile
    glutInitContextFlags(GLUT_DEBUG);          // Use OpenGL debug context

    // Initialize window
    glutInitWindowSize(win_width, win_height);
    glutCreateWindow("PV112 - cv2");

    // Initialize GLEW
    glewExperimental = GL_TRUE;
    glewInit();

    SetDebugCallback(simple_debug_callback);

    // Initialize out OpenGL stuff
    init();

    // Register callbacks
    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(key_pressed);
    glutMouseFunc(mouse_button_changed);
    glutMotionFunc(mouse_moved);
    glutTimerFunc(20, timer, 0);

    // Run the main loop
    glutMainLoop();

    return 0;
}
